from rdflib import URIRef, OWL, RDF, RDFS

from .owlapi_interface import OwlApiInterface


class FactPPError:
    """Postprocessing error after reasoning with FaCT++."""


class FaCTPPGraph:
    """Class for running the FaCT++ reasoner (using OwlApiInterface) and
    postprocessing the resulting inferred ontology.

    Parameters
    ----------
    graph : owlapi.Graph instance
        The graph to be inferred.
    """
    def __init__(self, graph):
        self.graph = graph
        self._inferred = None
        self._namespaces = None
        self._base_iri = None

    @property
    def inferred(self):
        """The current inferred graph."""
        if self._inferred is None:
            self._inferred = self.raw_inferred_graph()
        return self._inferred

    @property
    def base_iri(self):
        """Base iri of inferred ontology."""
        if self._base_iri is None:
            self._base_iri = URIRef(self.asserted_base_iri() + '-inferred')
        return self._base_iri

    @base_iri.setter
    def base_iri(self, value):
        """Assign inferred base iri."""
        self._base_iri = URIRef(value)

    @property
    def namespaces(self):
        """Namespaces defined in the original graph."""
        if self._namespaces is None:
            self._namespaces = dict(self.graph.namespaces()).copy()
            self._namespaces[''] = self.base_iri
        return self._namespaces

    def asserted_base_iri(self):
        """Returns the base iri or the original graph."""
        return URIRef(dict(self.graph.namespaces()).get('', '').rstrip('#/'))

    def raw_inferred_graph(self):
        """Returns the raw non-postprocessed inferred ontology as a rdflib
        graph."""
        return OwlApiInterface().reason(self.graph)

    def inferred_graph(self):
        """Returns the postprocessed inferred graph."""
        self.add_base_annotations()
        self.set_namespace()
        self.clean_base()
        self.remove_nothing_is_nothing()
        self.clean_ancestors()
        return self.inferred

    def add_base_annotations(self):
        """Copy base annotations from original graph to the inferred graph."""
        base = self.base_iri
        inferred = self.inferred
        for s, p, o in self.graph.triples(
                (self.asserted_base_iri(), None, None)):
            if p == OWL.versionIRI:
                version = o.rsplit('/', 1)[-1]
                o = URIRef('%s/%s' % (base, version))
            inferred.add((base, p, o))

    def set_namespace(self):
        """Override namespace of inferred graph with the namespace of the
        original graph.
        """
        inferred = self.inferred
        for k, v in self.namespaces.items():
            inferred.bind(k, v, override=True, replace=True)

    def clean_base(self):
        """Remove all relations `s? a owl:Ontology` where `s?` is not
        `base_iri`.
        """
        inferred = self.inferred
        for s, p, o in inferred.triples((None, RDF.type, OWL.Ontology)):
            inferred.remove((s, p, o))
        inferred.add((self.base_iri, RDF.type, OWL.Ontology))

    def remove_nothing_is_nothing(self):
        """Remove superfluid relation in inferred graph:

            owl:Nothing rdfs:subClassOf owl:Nothing
        """
        t = OWL.Nothing, RDFS.subClassOf, OWL.Nothing
        inferred = self.inferred
        if t in inferred:
            inferred.remove(t)

    def clean_ancestors(self):
        """Remove redundant rdfs:subClassOf relations in inferred graph."""
        inferred = self.inferred
        for s in inferred.subjects(RDF.type, OWL.Class):
            if isinstance(s, URIRef):
                parents = set(p for p in inferred.objects(s, RDFS.subClassOf)
                              if isinstance(p, URIRef))
                if len(parents) > 1:
                    for parent in parents:
                        ancestors = set(inferred.transitive_objects(
                            parent, RDFS.subClassOf))
                        for p in parents:
                            if p != parent and p in ancestors:
                                t = s, RDFS.subClassOf, p
                                if t in inferred:
                                    inferred.remove(t)



import collections

import rdflib
from rdflib import URIRef, RDF, RDFS, OWL

import owlready2
from owlready2 import World, Ontology, CURRENT_NAMESPACES
from owlready2.reasoning import (
    _apply_reasoning_results, _apply_inferred_obj_relations,
    _INFERRENCES_ONTOLOGY)

from .factppgraph import FaCTPPGraph


OWL_2_TYPE = {
    RDFS.subClassOf:        'class',
    RDFS.subPropertyOf:     'property',
    RDF.type:               'individual',
    OWL.equivalentClass:    'class',
    OWL.equivalentProperty: 'property',
}


def sync_reasoner_factpp(ontology_or_world=None, infer_property_values=False,
                         debug=1):
    """Run FaCT++ reasoner and load the inferred relations back into
    the owlready2 triplestore.

    Parameters
    ----------
    ontology_or_world : None | Ontology instance | World instance | list
        Identifies the world to run the reasoner over.
    infer_property_values : bool
        Whether to also infer property values.
    debug : bool
        Whether to print debug info to standard output.
    """
    if isinstance(ontology_or_world, World):
        world = ontology_or_world
    elif isinstance(ontology_or_world, Ontology):
        world = ontology_or_world.world
    elif isinstance(ontology_or_world, collections.Sequence):
        world = ontology_or_world[0].world
    else:
        world = owlready2.default_world

    if isinstance(ontology_or_world, Ontology):
        ontology = ontology_or_world
    elif CURRENT_NAMESPACES.get():
        ontology = CURRENT_NAMESPACES.get()[-1].ontology
    else:
        ontology = world.get_ontology(_INFERRENCES_ONTOLOGY)

    locked = world.graph.has_write_lock()
    if locked:
        world.graph.release_write_lock()  # Not needed during reasoning

    try:
        print('*** Prepare graph')
        # Exclude owl:imports because they are not needed and can
        # cause trouble when loading the inferred ontology
        g1 = rdflib.Graph()
        for s, p, o in world.as_rdflib_graph().triples((None, None, None)):
            if p != OWL.imports:
                g1.add((s, p, o))

        print('*** Run FaCT++ reasoner (and postprocess)')
        g2 = FaCTPPGraph(g1).inferred_graph()

        print('*** Load inferred ontology')
        # Check all rdfs:subClassOf relations in the inferred graph and add
        # them to the world if they are missing
        new_parents = collections.defaultdict(list)
        new_equivs = collections.defaultdict(list)
        entity_2_type = {}

        for s, p, o in g2.triples((None, None, None)):
            if (isinstance(s, URIRef) and
                    p in OWL_2_TYPE and
                    isinstance(o, URIRef)):
                s_storid = ontology._abbreviate(str(s), False)
                p_storid = ontology._abbreviate(str(p), False)
                o_storid = ontology._abbreviate(str(o), False)
                if (s_storid is not None and
                        p_storid is not None and
                        o_storid is not None):
                    if p in (RDFS.subClassOf, RDFS.subPropertyOf, RDF.type):
                        new_parents[s_storid].append(o_storid)
                        entity_2_type[s_storid] = OWL_2_TYPE[p]
                    else:
                        new_equivs[s_storid].append(o_storid)
                        entity_2_type[s_storid] = OWL_2_TYPE[p]

        if infer_property_values:
            inferred_obj_relations = []
            # Hmm, does FaCT++ infer any property values?
            # If not, remove the `infer_property_values` keyword argument.
            raise NotImplementedError

    finally:
        if locked:
            world.graph.acquire_write_lock()  # re-lock when applying results

    print('*** Applying reasoning results')
    _apply_reasoning_results(world, ontology, debug, new_parents, new_equivs,
                             entity_2_type)
    if infer_property_values:
        _apply_inferred_obj_relations(world, ontology, debug,
                                      inferred_obj_relations)



"""Python interface to the FaCT++ Reasoner.

This module is copied from the SimPhoNy project.

Original author: Matthias Urban
"""
import os
import subprocess
import logging
import rdflib
import tempfile
import argparse

logger = logging.getLogger(__name__)

RESULT_FILE = "_result_ontology.owl"


class OwlApiInterface():
    """Interface to the FaCT++ reasoner via OWLAPI."""

    def __init__(self):
        """Initialize the interface."""
        pass

    def reason(self, graph):
        """Generate the inferred axioms for a given Graph.

        Args:
            graph (Graph): An rdflib graph to execute the reasoner on.
        """
        with tempfile.NamedTemporaryFile("wt") as f:
            graph.serialize(f.name, format="xml")
            return self._run(f.name, command="--run-reasoner")

    def reason_files(self, *owl_files):
        """Merge the given owl and generate the inferred axioms.

        Args:
            owl_files (os.path): The owl files two merge
        """
        return self._run(*owl_files, command="--run-reasoner")

    def merge_files(self, *owl_files):
        """Merge the given owl files and its import closure.

        Args:
            owl_files (os.path): The owl files two merge
        """
        return self._run(*owl_files, command="--merge-only")

    def _run(self, *owl_files, command, output_file=None, return_graph=True):
        """Run the FaCT++ reasoner using a java command.

        Args:
            owl_files (str): Path to the owl files to load.
            command (str): Either --run-reasoner or --merge-only
            output_file (str, optional): Where the output should be stored.
                Defaults to None.
            return_graph (bool, optional): Whether the result should be parsed
                and returned. Defaults to True.

        Returns:
            rdflib.Graph: The reasoned result.
        """
        java_base = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "java")
        )
        cmd = [
            "java", "-cp",
            java_base + "/lib/jars/*",
            "-Djava.library.path="
            + java_base + "/lib/so", "org.simphony.OntologyLoader"
        ] + ["%s" % command] + list(owl_files)
        logger.info("Running Reasoner")
        logger.debug(f"Command {cmd}")
        subprocess.run(cmd, check=True)

        graph = None
        if return_graph:
            graph = rdflib.Graph()
            graph.parse(RESULT_FILE)
        if output_file:
            os.rename(RESULT_FILE, output_file)
        else:
            os.remove(RESULT_FILE)
        return graph


def reason_from_terminal():
    """Run the reasoner from terminal."""
    parser = argparse.ArgumentParser(
        description="Run the FaCT++ reasoner on the given OWL file. "
        "Catalog files are used to load the import closure. "
        "Then the reasoner is executed and the inferred triples are merged "
        "with the asserted ones. If multiple OWL files are given, they are "
        "merged beforehand"
    )
    parser.add_argument("owl_file", nargs="+",
                        help="OWL file(s) to run the reasoner on.")
    parser.add_argument("output_file",
                        help="Path to store inferred axioms to.")

    args = parser.parse_args()
    OwlApiInterface()._run(*args.owl_file, command="--run-reasoner",
                           return_graph=False, output_file=args.output_file)


if __name__ == "__main__":
    reason_from_terminal()
